#include <boost/filesystem.hpp>
#include <boost/algorithm/string.hpp>
#include <iostream>
#include <algorithm>
#include "filesystem.h"


bool supportedImage (const boost::filesystem::path & p) {
    static const std::string supportedExts[] = {".tif", ".png", ".bmp", ".jpg"};
    const std::string ext = p.extension().generic_string();
    const std::string lowerext = boost::algorithm::to_lower_copy(ext);

    const std::string * supportedExtsBegin = supportedExts;
    const std::string * supportedExtsEnd   = supportedExts + (sizeof(supportedExts) / sizeof(std::string));
    return (supportedExtsEnd != std::find(supportedExtsBegin, supportedExtsEnd, lowerext));
}

bool notSupportedImage (const boost::filesystem::path & p) {
    return !supportedImage(p);
}

enum STR2INT_ERROR { SUCCESS, OVERFLOW, UNDERFLOW, INCONVERTIBLE };

STR2INT_ERROR str2int (int &i, char const *s, int base = 0)
{
    char *end;
    long  l;
    errno = 0;
    l = strtol(s, &end, base);
    if ((errno == ERANGE && l == LONG_MAX) || l > INT_MAX) {
        return OVERFLOW;
    }
    if ((errno == ERANGE && l == LONG_MIN) || l < INT_MIN) {
        return UNDERFLOW;
    }
    if (*s == '\0' || *end != '\0') {
        return INCONVERTIBLE;
    }
    i = l;
    return SUCCESS;
}

bool compareNumberedStrings (const std::string & lhs, const std::string & rhs) {
    unsigned int commonSuffixLength = 0;
    for (unsigned int i = 1; i <= std::min(lhs.size(), rhs.size()); ++i) {
        if (std::equal(lhs.rbegin(), lhs.rbegin() + i, rhs.rbegin())) {
            ++commonSuffixLength;
        } else {
            break;
        }
    }
    unsigned int commonPrefixLength = 0;
    std::string specificLhs, specificRhs;
    int numberLhs, numberRhs;
    for (unsigned int i = 1; i <= std::min(lhs.size() - commonSuffixLength, rhs.size() - commonSuffixLength); ++i) {
        specificLhs = lhs.substr(i - 1, lhs.size() - commonSuffixLength - i + 1);
        specificRhs = rhs.substr(i - 1, rhs.size() - commonSuffixLength - i + 1);
        /// We include one more character in the common prefix iff we have not reached the part where different numbers begin
        /// If we do not do this, cases like "asdf-107.tif" vs. "asdf-1003.tif" break into "asdf-10|7|.tif" and "asdf-10|03|.tif",
        /// which would result in an incorrect comparison.
        if ((SUCCESS == str2int(numberLhs, specificLhs.c_str(), 10)) && (SUCCESS == str2int(numberRhs, specificRhs.c_str(), 10)) && (numberLhs != numberRhs)) {
            if ((0 > numberLhs) || (0 > numberRhs)) {
                /// There is a hyphen before the number, but we don't want to interpret it as a number,
                /// so we advance one character. The || instead of && is because one of them could
                /// be zero.
            } else {
                break;
            }
        } else {
            if (std::equal(lhs.begin(), lhs.begin() + i, rhs.begin())) {
                ++commonPrefixLength;
            } else {
                break;
            }
        }
    }
//    std::cout << "Common prefix length: " << commonPrefixLength << "; "
//              << lhs.substr(0, commonPrefixLength) << " <-> " << rhs.substr(0, commonPrefixLength) << std::endl;
//    std::cout << "Common prefix length: " << commonSuffixLength << "; "
//              << lhs.substr(lhs.size() - commonSuffixLength) << " <-> " << rhs.substr(rhs.size() - commonSuffixLength) << std::endl;
//    std::cout << "Different part: " << specificLhs << " <-> " << specificRhs << std::endl;
    if ((SUCCESS == str2int(numberLhs, specificLhs.c_str(), 10)) && (SUCCESS == str2int(numberRhs, specificRhs.c_str(), 10)) && (numberLhs != numberRhs)) {
        /// If we are here, it's because we have determined a common (possibly null) prefix and suffix for both operands
        /// and whatever is left between them is only a (different) number in each operand. I am not sure that the provision
        /// that the numbers are different is necessary at this point because it was already checked before, but I leave it for now.
        return numberLhs < numberRhs;
    } else {
        return specificLhs < specificRhs;
    }
}

// This function does the same as operator< but instead of using lexicographic comparison
// it compares so that xxx_yyy_23 < xxx_yyy_132, which is not true with lexicographic
// comparisons.
bool comparePaths (const boost::filesystem::path & lhs, const boost::filesystem::path & rhs) {
    return compareNumberedStrings(lhs.stem().generic_string(), rhs.stem().generic_string());
}


PathVector getImageList(const std::string & pathCandidate) {
    PathVector candidateImagePaths;
    try {
        assert (boost::filesystem::exists(pathCandidate) && boost::filesystem::is_directory(pathCandidate));    // does p actually exist?
        std::remove_copy_if(boost::filesystem::directory_iterator(pathCandidate), boost::filesystem::directory_iterator(), std::back_inserter(candidateImagePaths), notSupportedImage);
        std::sort(candidateImagePaths.begin(), candidateImagePaths.end(), comparePaths);             // sort, since directory iteration
    } catch (const boost::filesystem::filesystem_error& ex) {
        std::cerr << ex.what() << '\n';
    }
    return candidateImagePaths;
}
