#ifndef THUMBNAIL_H
#define THUMBNAIL_H

#include "tinymt32.h"

typedef struct CUstream_st *cudaStream_t;



class Thumbnail {
public:
    enum State {NOT_ANALYSED = 0,
                TO_ANALYSE   = 1,
                NEGATIVE     = 2,
                POSITIVE     = 3};

    Thumbnail ();

    Thumbnail (unsigned int width, unsigned int height, unsigned int seed = 666);

    virtual ~Thumbnail ();

    /// Sets every pixel to NOT_ANALYSED. Does nothing if the thumbnail is disabled.
    void clear ();

    unsigned char & elementAt (int x, int y);

    /// Sets every POSITIVE pixel to TO_ANALYSE and the rest to NOT_ANALYSED.
    /// Does nothing if the thumbnail is disabled.
    void initializeNextFrame ();

    /// Set blocksPerTile random pixels to TO_ANALYSE for each tile of specified
    /// dimensions. Does nothing if the thumbnail is disabled.
    void addRandom (unsigned int blocksPerTile,
                    unsigned int tileWidth,
                    unsigned int tileHeight);

    /// Prepares the thumbnail for the next iteration, marking for analysis
    /// pixels next to any POSITIVE. Returns true iff a next iteration is
    /// necessary. If the thumbnail is disabled it always returns false.
    bool nextIteration (int radius = 1);

    /// Copies host data to device iff the thumbnail is enabled.
    void toDevice (cudaStream_t stream = NULL) const;

    /// Copies device data to host iff the thumbnail is enabled.
    void fromDevice (cudaStream_t stream = NULL);

    unsigned int getWidth () const;

    unsigned int getHeight () const;

    unsigned char const * getHostData () const;
    unsigned char       * getHostData ();

    unsigned char       * getDeviceData ();

    bool getEnabled () const;
    void setEnabled (bool value);

    void allocate (unsigned int width, unsigned int height);
    void release();
    void setRandomSeed(unsigned int seed);

    void advanceFrame() { this->frame++; this->iteration = 0; }

    bool writeImage();

private:
    Thumbnail (const Thumbnail & other);

    unsigned int width;
    unsigned int height;
    unsigned char *hostData;
    unsigned char *auxHostData; // For nextIteration mainly.
    unsigned char *deviceData;

    tinymt32_t rngState;

    bool enabled;

    unsigned int frame;
    unsigned int iteration;
};

#endif // THUMBNAIL_H
